/*
 * =====================================================================================
 *
 *       Filename:  v4l2_capture.c
 *
 *    Description:  v4l2_capture c file
 *
 *        Version:  1.0
 *
 *         Author:  Alexander Aring, a.aring@phytec.de
 *        Company:  PHYTEC
 *
 * =====================================================================================
 */

#include "v4l2_capture.h"

static int fd = -1;
static struct buffer *buffers = NULL;
static __u32 n_buffers = 0;

/* TODO parameter */
extern char *register_timeout_file_path;
/* TODO prototype */
int write_cam_reg(__u8 address, __u8 reg,
		__u16 val, char mode);


/*-----------------------------------------------------------------------------
 *  Set a value to a camera register.
 *-----------------------------------------------------------------------------*/
int set_cam_reg(__u32 address, __u64 reg, __u64 data)
{
	int err;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	regmatch.type = V4L2_CHIP_MATCH_I2C_ADDR;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	camreg.val = data;

	err = ioctl(fd, VIDIOC_DBG_S_REGISTER, &camreg);
	if (err < 0) {
		perror("VIDIOC_DBG_S_REGISTER");
		return -1;
	}
	
	return 0;
}

/*-----------------------------------------------------------------------------
 *  Get a value to a camera register.
 *-----------------------------------------------------------------------------*/
int get_cam_reg(__u32 address, __u64 reg, __u64 *data)
{
	int err;
	struct v4l2_dbg_register camreg;
	struct v4l2_dbg_match regmatch;

	regmatch.type = V4L2_CHIP_MATCH_I2C_ADDR;
	regmatch.addr = address;

	camreg.match = regmatch;
	camreg.size = REGISTER_SIZE;
	camreg.reg = reg;
	
	err = ioctl(fd, VIDIOC_DBG_G_REGISTER, &camreg);
	if (err < 0) {
		perror("VIDIOC_DBG_G_REGISTER");
		return -1;
	}

	*data = camreg.val;

	return 0;
}

int open_video_device(char *dev_name)
{
	int err = -1;
	struct stat st;

	err = stat(dev_name, &st);
	if (err < 0) {
		perror(dev_name);
		return -1;
	}

	if (!S_ISCHR(st.st_mode)) {
		fprintf(stderr, "%s is no device\n", dev_name);
		return -1;
	}

	fd = open(dev_name, O_RDWR, 0);
	if (fd == -1) {
		perror("open");
		return -1;
	}

	return 0;
}

int close_video_device()
{
	int err;

	err = close(fd);
	if (err == -1) {
		perror("close");
		return -1;
	}

	fd = -1;
	return 0;
}

int init_video_device(char *dev_name)
{
	int err;
	v4l2_std_id std_id;
	struct v4l2_capability cap;

	err = ioctl(fd, VIDIOC_QUERYCAP, &cap);
	if (err < 0) {
		if (errno == EINVAL)
			fprintf(stderr, "%s is no V4L2 device\n",
					dev_name);
		else
			perror("VIDIOC_QUERYCAP");
		return -1;
	}

	if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
		fprintf(stderr, "%s is no video capture device\n",
				dev_name);
		return -1;
	}

	if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
		fprintf(stderr, "%s doesn't support mmap as IO method\n",
				dev_name);
		return -1;
	}

	/* Setting video standard to pal B/G
	 * if device supports that */
	std_id = V4L2_STD_PAL_BG;
	
	err = ioctl(fd, VIDIOC_S_STD, &std_id);
	if (err < 0) {
		/* EINVAL returns, when device
		 * don't support this video strandard */
		if (errno != EINVAL) {
			perror ("VIDIOC_S_STD");
			return -1;
		}
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Get max framesize with VIDIOC_TRY_FMT with UINT_MAX for width and height.
 *  Hope that will always return the max framesize for pixel_format.
 *-----------------------------------------------------------------------------*/
int get_max_framesize(__u32 *width, __u32 *height,
		__u32 pixel_format)
{
	int err;
	struct v4l2_format format;

	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	format.fmt.pix.width = UINT_MAX;
	format.fmt.pix.height = UINT_MAX;
	format.fmt.pix.pixelformat = pixel_format;
	format.fmt.pix.field = V4L2_FIELD_ANY;

	err = ioctl(fd, VIDIOC_TRY_FMT, &format);
	if (err < 0) {
		perror("VIDIOC_TRY_FMT");
		return -1;
	}

	err = ioctl(fd, VIDIOC_G_FMT, &format);
	if (err < 0) {
		perror("VIDIOC_G_FMT");
		return -1;
	}

	*width = format.fmt.pix.width;
	*height = format.fmt.pix.height;

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Set a format with height and width by pixel_format.
 *  Return imagesize by call by reference.
 *-----------------------------------------------------------------------------*/
static int set_format(__u32 width, __u32 height,
		__u32 pixel_format, __u32 *imagesize)
{
	int err;
	struct v4l2_format format;

	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	format.fmt.pix.width = width;
	format.fmt.pix.height = height;
	format.fmt.pix.pixelformat = pixel_format;
	format.fmt.pix.field = V4L2_FIELD_ANY;

	err = ioctl(fd, VIDIOC_S_FMT, &format);
	if (err < 0) {
		perror("VIDIOC_S_FMT");
		return -1;
	}

	err = ioctl(fd, VIDIOC_G_FMT, &format);
	if (err < 0) {
		perror("VIDIOC_G_FMT");
		return -1;
	}

	*imagesize = format.fmt.pix.sizeimage;

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Initialize mmap for buffers.
 *-----------------------------------------------------------------------------*/
static int init_mmap()
{
	int err;

	struct v4l2_requestbuffers req;

	memset(&req, 0, sizeof(struct v4l2_requestbuffers));

	req.count = BUFFER_COUNT;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;

	err = ioctl(fd, VIDIOC_REQBUFS, &req);
	if (err < 0) {
		perror("VIDIOC_REQBUFS");
		return -1;
	}
	
	buffers = malloc(sizeof(struct buffer) * req.count);
	if (!buffers) {
		fprintf (stderr, "Out of memory\n");
		return -1;
	}

	for (n_buffers=0; n_buffers < req.count; ++n_buffers) {
		struct v4l2_buffer buf;

		memset(&buf, 0, sizeof(struct v4l2_buffer));

		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = n_buffers;

		err = ioctl(fd, VIDIOC_QUERYBUF, &buf);
		if (err < 0) {
			perror("VIDIOC_QUERYBUF");
			goto out;
		}

		buffers[n_buffers].length = buf.length;
		buffers[n_buffers].start =
			mmap(NULL,
					buf.length,
					PROT_READ | PROT_WRITE,
					MAP_SHARED,
					fd, buf.m.offset);

		if (buffers[n_buffers].start == MAP_FAILED) {
			perror("mmap");
			goto out;
		}
	}

	return 0;
out:
	free(buffers);
	return -1;
}

static int start_capturing()
{
	int err;
	unsigned int i;

	enum v4l2_buf_type type;

	for (i=0; i < n_buffers; ++i) {
		struct v4l2_buffer buf;

		memset(&buf, 0, sizeof(struct v4l2_buffer));

		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = i;

		err = ioctl(fd, VIDIOC_QBUF, &buf);
		if (err < 0) {
			perror("VIDIOC_QBUF");
			return -1;
		}
	}

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	err = ioctl(fd, VIDIOC_STREAMON, &type);
	if (err < 0) {
		perror("VIDIOC_STREAMON");
		return -1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Stop capture of camera.
 *-----------------------------------------------------------------------------*/
static int stop_capturing()
{
	int err;
	enum v4l2_buf_type type;

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	err = ioctl(fd, VIDIOC_STREAMOFF, &type);
	if (err == -1) {
		perror("VIDIOC_STREAMOFF");
		return -1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Read a frame and call process_frame.
 *-----------------------------------------------------------------------------*/
static int read_frame(int process_frame(const void *image, __u32 imagesize), __u32 imagesize, unsigned int count)
{
	struct v4l2_buffer buf;
	int err;

	memset(&buf, 0, sizeof(struct v4l2_buffer));

	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;

	err = ioctl(fd, VIDIOC_DQBUF, &buf);
	if (err == -1) {
		if (errno == EAGAIN) {
			return ERR_IMAGE_AGAIN;
		} else {
			perror("VIDIOC_QBUF");
			return ERR_IMAGE_ERROR;
		}
	}

	assert(buf.index < n_buffers);

	if (count == 1) {
		err = process_frame(buffers[buf.index].start, imagesize);
		if (err < 0) {
			fprintf(stderr, "Picture capture failed!\n");
			return ERR_IMAGE_ERROR;
		}
	}

	err = ioctl(fd, VIDIOC_QBUF, &buf);
	if (err == -1) {
		perror("VIDIOC_QBUF");
		return ERR_IMAGE_ERROR;
	}

	return ERR_IMAGE_OK;
}

/*-----------------------------------------------------------------------------
 *  Capture a count of images.
 *-----------------------------------------------------------------------------*/
static int record_image(unsigned int timeout, unsigned int count,
		int process_frame(const void *image, __u32 imagesize), __u32 imagesize)
{
	fd_set fds;
	int err;
	struct timeval tv;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	while (count) {
		for (;;) {
			FD_ZERO (&fds);
			FD_SET (fd, &fds);

			if (timeout)
				err = select(fd + 1, &fds, NULL, NULL, &tv);
			else
				err = select(fd + 1, &fds, NULL, NULL, NULL);

			if (err == -1) {
				perror("select");
				return -1;
			}
			if (err == 0) {
				fprintf(stderr, "timeout\n");
				return -1;
			}

			err = read_frame(process_frame, imagesize, count);
			if (err ==  ERR_IMAGE_AGAIN) {
				fprintf(stderr, "Error capture image, retrying...");
				continue;
			} else if (err == ERR_IMAGE_OK) {
				break;
			} else {
				return -1;
			}
		}
		count--;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Uninitialize the mmap from buffers.
 *-----------------------------------------------------------------------------*/
static int uninit_mmap()
{
	int err;
	unsigned int i;

	for (i = 0; i < n_buffers; i++) {
		err = munmap(buffers[i].start, buffers[i].length);
		if (err < 0) {
			perror("munmap");
			return -1;
		}
	}

	free(buffers);

	return 0;
}

/*-----------------------------------------------------------------------------
 *  Capture a image with width, height and pixel_format. Then call process_frame.
 *-----------------------------------------------------------------------------*/
int capture_image(unsigned int timeout,
		unsigned int count,
		__u32 width, 
		__u32 height, 
		__u32 pixel_format,
		int process_frame(const void *image, __u32 imagesize))
{
	int err;
	__u32 imagesize;

	err = set_format(width, height,
			pixel_format, &imagesize);
	if (err < 0)
		return -1;

	err = init_mmap();
	if (err < 0)
		return -1;

	err = start_capturing();
	if (err < 0)
		goto err_mmap;

	err = record_image(timeout, count,
			process_frame, imagesize);
	if (err < 0)
		goto err_cap;

	err = stop_capturing();
	if (err < 0)
		goto err_mmap;

	err = uninit_mmap();
	if (err < 0)
		return -1;

	return 0;
err_cap:
	if (register_timeout_file_path) {
		err = do_reg_file_parse(register_timeout_file_path, write_cam_reg);
		if (err <0) {
			fprintf(stderr, "Failed to parse register file on timeout!\n");
		}
	}
	stop_capturing();
err_mmap:
	uninit_mmap();
	return -1;
}

